package chocopy.pa3;

import java.util.*;

import chocopy.common.analysis.SymbolTable;
import chocopy.common.analysis.AbstractNodeAnalyzer;
import chocopy.common.analysis.types.ClassValueType;
import chocopy.common.analysis.types.ListValueType;
import chocopy.common.analysis.types.SymbolType;
import chocopy.common.analysis.types.ValueType;
import chocopy.common.astnodes.*;
import chocopy.common.codegen.*;

import static chocopy.common.codegen.RiscVBackend.Register.*;

/**
 * This is where the main implementation of PA3 will live.
 *
 * A large part of the functionality has already been implemented
 * in the base class, CodeGenBase. Make sure to read through that
 * class, since you will want to use many of its fields
 * and utility methods in this class when emitting code.
 *
 * Also read the PDF spec for details on what the base class does and
 * what APIs it exposes for its sub-class (this one). Of particular
 * importance is knowing what all the SymbolInfo classes contain.
 */
public class CodeGenImpl extends CodeGenBase {
    private class BetterRsicVBackend {
        private final RiscVBackend backend;

        BetterRsicVBackend(RiscVBackend _backend) {
            backend = _backend;
        }

        public void emitNoop(String comment) {
            if (comment != null) {
                backend.emitMV(ZERO, ZERO, "Noop: " + comment);
            }
        }

        public void emitPush(RiscVBackend.Register reg, String comment) {
            if (comment == null) comment = "";
            comment = "PUSH: " + comment;
            backend.emitADDI(SP, SP, -1 * backend.getWordSize(), comment);
            backend.emitSW(reg, SP, 0, comment);
        }

        public void emitPop(RiscVBackend.Register reg, String comment) {
            if (comment == null)
                comment = "";
            comment = "POP: " + comment;
            if (reg != null)
                backend.emitLW(reg, SP, 0, comment);
            backend.emitADDI(SP, SP, backend.getWordSize(), comment);
        }

        public void emitCall(Label calledLabel, String comment, boolean isRecursive, boolean isBuiltIn) {
            // Arguments should be already pushed to stack.

            if(isBuiltIn) {
                // Don't set LOGICAL parent FP for built-in function.
                backend.emitJAL(calledLabel, comment);
                return;
            }

            if(isRecursive) {
                // LOGICAL parent FP = parent's LOGICAL parent FP
                backend.emitLW(T0, FP, 0, "emitCall: save LOGICAL parent FP as last argument - recursive. 0(new_fp)");
                emitPush(T0, "emitCall: save LOGICAL parent FP as last argument - recursive. 0(new_fp)");
            }
            else {
                // LOGICAL parent FP = RUNTIME parent FP
                emitPush(FP, "emitCall: save LOGICAL parent FP as last argument - Non-recursive. 0(new_fp)");
            }

            backend.emitJAL(calledLabel, comment);

            emitPop(null, "emitCall: throw LOGICAL parent FP to dustbin.");
        }

        public void emitCallR(RiscVBackend.Register calledRegister, String comment, boolean isRecursive, boolean isBuiltIn) {
            // Arguments should be already pushed to stack.

            if(isBuiltIn) {
                // Don't set LOGICAL parent FP for built-in function.
                backend.emitJALR(calledRegister, comment);
                return;
            }

            if(isRecursive) {
                // LOGICAL parent FP = parent's LOGICAL parent FP
                backend.emitLW(T0, FP, 0, "emitCall: save LOGICAL parent FP as last argument - recursive. 0(new_fp)");
                emitPush(T0, "emitCall: save LOGICAL parent FP as last argument - recursive. 0(new_fp)");
            }
            else {
                // LOGICAL parent FP = RUNTIME parent FP
                emitPush(FP, "emitCall: save LOGICAL parent FP as last argument - Non-recursive. 0(new_fp)");
            }

            backend.emitJALR(calledRegister, comment);

            emitPop(null, "emitCall: throw LOGICAL parent FP to dustbin.");
        }


        public void emitFunctionBegin(String funcName, String comment) {
            emitNoop(comment);
            emitPush(RA, "backup my return address. -4(new_fp)");
            emitPush(FP, "backup RUNTIME parent fp. -8(new_fp)");
            backend.emitADDI(FP, SP, 2 * backend.getWordSize(), "Change FP to apply new frame.");

            backend.emitJ(new Label("$" + funcName + "$saveReg"), "Jump to save callee-saved registers");
            backend.emitLocalLabel(new Label("$" + funcName + "$saveRegContinue"), "Begin real function codes:");
        }

        public void emitFunctionEnd(String funcName, Void regMgr, String comment) {
            // Return value should be in A0
            List<RiscVBackend.Register> savedRegs = new ArrayList<RiscVBackend.Register>() {}; // regMgr.registerToSaveAndRestoreInFunc;

            // Function return point:
            backend.emitLocalLabel(new Label("$" + funcName + "$exitPoint"), "Function return point");

            emitNoop(comment);
            backend.emitADDI(SP, FP, -(2 + savedRegs.size()) * backend.getWordSize(), "Revert all local variables on this dying frame.");
            for (int i = savedRegs.size(); i-- > 0; ) {
                emitPop(savedRegs.get(i), "restore callee-saved reg.");
            }
            emitPop(FP, "restore RUNTIME parent FP.");
            emitPop(RA, "restore my return address.");
            backend.emitJR(RA, "Real return!");

            backend.emitLocalLabel(new Label("$" + funcName + "$saveReg"), "Save callee-saved registers before use.");
            for (int i = 0; i < savedRegs.size(); ++i) {
                emitPush(savedRegs.get(i), "save callee-saved reg.");
            }
            backend.emitJ(new Label("$" + funcName + "$saveRegContinue"), "go back to execute real function code.");
        }

        public void emitFunctionReturn(String funcName, String comment) {
            backend.emitJ(new Label("$" + funcName + "$exitPoint"), comment);
        }

        public void emitWrapIntValToIntObject(RiscVBackend.Register tmpReg, SymbolType typeIntOrBool) {
            emitPop(tmpReg, "INT VAL TO INT OBJ: POP");
            emitAllocPrototype(typeIntOrBool, tmpReg);
            emitPush(A0, "INT VAL TO INT OBJ: PUSH");
        }

        public void emitMemcpy(RiscVBackend.Register dstAddr_nonconst, RiscVBackend.Register srcAddr_nonconst, RiscVBackend.Register sizeInWord_nonconst, RiscVBackend.Register tmpReg, String labelPrefix) {
            Label loopBegin = new Label(labelPrefix + ".memCpy.loopBegin");
            Label loopExit = new Label(labelPrefix + ".memCpy.loopExit");
            backend.emitLocalLabel(loopBegin, "memcpy");
            backend.emitBEQ(sizeInWord_nonconst, ZERO, loopExit, "If wordLeft is zero, exit.");
            backend.emitLW(tmpReg, srcAddr_nonconst, 0, "LW");
            backend.emitSW(tmpReg, dstAddr_nonconst, 0, "SW");
            backend.emitADDI(srcAddr_nonconst, srcAddr_nonconst, backend.getWordSize(), "INC srcAddr");
            backend.emitADDI(dstAddr_nonconst, dstAddr_nonconst, backend.getWordSize(), "INC dstAddr");
            backend.emitADDI(sizeInWord_nonconst, sizeInWord_nonconst, -1, "DEC sizeInWord");
            backend.emitJ(loopBegin, "goto begin");
            backend.emitLocalLabel(loopExit, "memcpy");
        }

        public void emitAllocPrototype(SymbolType typ, RiscVBackend.Register objReg) {
            // Convention: if typ is INT_TYPE or BOOL_TYPE then objReg contains the literal value, not a pointer.
            //             else, objReg contains the object length to allocate.
            // @returns A0

            Label prototypeLabel = objectClass.getPrototypeLabel();
            if (typ.isListType() || typ.equals(SymbolType.EMPTY_TYPE))
                prototypeLabel = listClass.getPrototypeLabel();
            else if (typ.equals(SymbolType.INT_TYPE))
                prototypeLabel = intClass.getPrototypeLabel();
            else if (typ.equals(SymbolType.STR_TYPE))
                prototypeLabel = strClass.getPrototypeLabel();
            else if (typ.equals(SymbolType.BOOL_TYPE))
                prototypeLabel = boolClass.getPrototypeLabel();

            backend.emitLA(A0, prototypeLabel, "EMIT_ALLOC: Load prototype address.");

            if (typ.equals(SymbolType.INT_TYPE) || typ.equals(SymbolType.BOOL_TYPE)) {
                assert objReg != T0 && objReg != T1 && objReg != T2 && objReg != T3;
                backend.emitJAL(objectAllocLabel, "EMIT_ALLOC: Allocate int/bool");
                backend.emitSW(objReg, A0, 3 * backend.getWordSize(), "EMIT_ALLOC: Set __int__ or __bool__ value.");
            }
            else {
                backend.emitMV(A1, objReg, "EMIT_ALLOC: Move size to A1.");
                backend.emitJAL(objectAllocResizeLabel, "EMIT_ALLOC: Allocate raw memory");
            }
        }

        private void __rlib_emitPushIntVal(RiscVBackend.Register val_nonconst, String comment) {
            emitNoop(comment);
            emitPush(val_nonconst, "PUSH_INTOBJ_TO_STACK (REAL VAL)");
            backend.emitLA(val_nonconst, intClass.getDispatchTableLabel(), "OBJECT HEAD - DISPATCH TBL");
            emitPush(val_nonconst, "PUSH_INTOBJ_TO_STACK");
            backend.emitLI(val_nonconst, 4, "OBJECT HEAD - SIZE = 3+1");
            emitPush(val_nonconst, "PUSH_INTOBJ_TO_STACK");
            backend.emitLI(val_nonconst, 1, "OBJECT HEAD - TYPE = INT");
            emitPush(val_nonconst, "PUSH_INTOBJ_TO_STACK");
            backend.emitMV(val_nonconst, SP, "Return INT address.");
        }
        public void emitDebugPrintReg_withoutBackupEverything(RiscVBackend.Register regToPrint_const) {
            __rlib_emitPushIntVal(regToPrint_const, "DEBUG_KIT PRINT_REG: push reg (and backup reg)");
            emitPush(regToPrint_const, "DEBUG_KIT PRINT_REG: push arg to print");
            backend.emitJAL(new Label("$print"), "DEBUG_KIT PRINT_REG: do print");
            emitPop(null, "DEBUG_KIT PRINT_REG: cleanup");
            emitPop(null, "DEBUG_KIT PRINT_REG: cleanup");
            emitPop(null, "DEBUG_KIT PRINT_REG: cleanup");
            emitPop(null, "DEBUG_KIT PRINT_REG: cleanup");
            emitPop(regToPrint_const, "DEBUG_KIT PRINT_REG: restore reg");
        }

        public void emitDebugPrintReg(RiscVBackend.Register regToPrint_const) {
            emitPush(T0, "DEBUG_KIT PRINT_REG: BACKUP EVERYTHING");
            emitPush(T1, "DEBUG_KIT PRINT_REG: BACKUP EVERYTHING");
            emitPush(A0, "DEBUG_KIT PRINT_REG: BACKUP EVERYTHING");
            emitPush(A1, "DEBUG_KIT PRINT_REG: BACKUP EVERYTHING");
            emitDebugPrintReg_withoutBackupEverything(regToPrint_const);
            emitPop(A1, "DEBUG_KIT PRINT_REG: RESTORE EVERYTHING");
            emitPop(A0, "DEBUG_KIT PRINT_REG: RESTORE EVERYTHING");
            emitPop(T1, "DEBUG_KIT PRINT_REG: RESTORE EVERYTHING");
            emitPop(T0, "DEBUG_KIT PRINT_REG: RESTORE EVERYTHING");
        }
        public void emitDebugAbort() {
            backend.emitJ(new Label("abort"), "DEBUG_KIT ABORT");
        }

    }

    /** A code generator emitting instructions to BACKEND. */
    public CodeGenImpl(RiscVBackend backend) {
        super(backend);
        betterBackend = new BetterRsicVBackend(backend);
    }

    private BetterRsicVBackend betterBackend;

    /** Operation on None. */
    private final Label errorNone = new Label("error.None");
    /** Division by zero. */
    private final Label errorDiv = new Label("error.Div");
    /** Index out of bounds. */
    private final Label errorOob = new Label("error.OOB");

    private Map<FuncInfo, StmtAnalyzer> mapFuncInfoToStmtAnalyzer = new HashMap<>();

    /**
     * Emits the top level of the program.
     *
     * This method is invoked exactly once, and is surrounded
     * by some boilerplate code that: (1) initializes the heap
     * before the top-level begins and (2) exits after the top-level
     * ends.
     *
     * You only need to generate code for statements.
     *
     * @param statements top level statements
     */
    protected void emitTopLevel(List<Stmt> statements) {
        StmtAnalyzer stmtAnalyzer = new StmtAnalyzer(null);
        backend.emitADDI(SP, SP, -2 * backend.getWordSize(),
                         "Saved FP and saved RA (unused at top level).");
        backend.emitSW(ZERO, SP, 0, "Top saved FP is 0.");
        backend.emitSW(ZERO, SP, 4, "Top saved RA is 0.");
        backend.emitADDI(FP, SP, 2 * backend.getWordSize(),
                         "Set FP to previous SP.");
        for (Stmt stmt : statements) {
            stmt.dispatch(stmtAnalyzer);
        }
        backend.emitLI(A0, EXIT_ECALL, "Code for ecall: exit");
        backend.emitEcall(null);
    }

    /**
     * Emits the code for a function described by FUNCINFO.
     *
     * This method is invoked once per function and method definition.
     * At the code generation stage, nested functions are emitted as
     * separate functions of their own. So if function `bar` is nested within
     * function `foo`, you only emit `foo`'s code for `foo` and only emit
     * `bar`'s code for `bar`.
     */
    protected void emitUserDefinedFunction(FuncInfo funcInfo) {
        backend.emitGlobalLabel(funcInfo.getCodeLabel());
        StmtAnalyzer stmtAnalyzer;
        // LocalTable(analyzer) already initialized by this function's child.
        if (mapFuncInfoToStmtAnalyzer.containsKey(funcInfo))
            stmtAnalyzer = mapFuncInfoToStmtAnalyzer.get(funcInfo);
        else
            stmtAnalyzer = new StmtAnalyzer(funcInfo);

        betterBackend.emitFunctionBegin(funcInfo.getFuncName(), null);
        stmtAnalyzer.pushLocalVars(funcInfo.getLocals()); // regMgr should NEVER be used, because of this statement.

        // Emit code for function stmts
        for (Stmt stmt : funcInfo.getStatements()) {
            stmt.dispatch(stmtAnalyzer);
        }

        betterBackend.emitFunctionEnd(funcInfo.getFuncName(), null, null);
    }

    /** An analyzer that encapsulates code generation for statments. */
    private class StmtAnalyzer extends AbstractNodeAnalyzer<Void> {
        /*
         * The symbol table has all the info you need to determine
         * what a given identifier 'x' in the current scope is. You can
         * use it as follows:
         *   SymbolInfo x = sym.get("x");
         *
         * A SymbolInfo can be one the following:
         * - ClassInfo: a descriptor for classes
         * - FuncInfo: a descriptor for functions/methods
         * - AttrInfo: a descriptor for attributes
         * - GlobalVarInfo: a descriptor for global variables
         * - StackVarInfo: a descriptor for variables allocated on the stack,
         *      such as locals and parameters
         *
         * Since the input program is assumed to be semantically
         * valid and well-typed at this stage, you can always assume that
         * the symbol table contains valid information. For example, in
         * an expression `foo()` you KNOW that sym.get("foo") will either be
         * a FuncInfo or ClassInfo, but not any of the other infos
         * and never null.
         *
         * The symbol table in funcInfo has already been populated in
         * the base class: CodeGenBase. You do not need to add anything to
         * the symbol table. Simply query it with an identifier name to
         * get a descriptor for a function, class, variable, etc.
         *
         * The symbol table also maps nonlocal and global vars, so you
         * only need to lookup one symbol table and it will fetch the
         * appropriate info for the var that is currently in scope.
         */

        /** Symbol table for my statements. */
        private SymbolTable<SymbolInfo> sym;

        /** Label of code that exits from procedure. */
        protected Label epilogue;

        /** The descriptor for the current function, or null at the top
         *  level. */
        private FuncInfo funcInfo;

        private int _r_local_label_counter = 427;

        // Extend StackVarInfo to add more info.
        private Map<String, Integer> localMap = new HashMap<>(); // Map from StackVarInfo to offset of FP.
        private void initArgsMap(List<String> args) {
            // 0(fp) is LOGICAL parent FP now. so arg0 is at 4(fp).
            for (int i = 0; i < args.size(); ++i) {
                int offset = (args.size() - i /* -1 */) * backend.getWordSize();
                localMap.put(args.get(i), offset);
            }
        }
        private void initLocalVars(List<StackVarInfo> locals, int firstLocalOffset) {
            for (int i = 0; i < locals.size(); ++i) {
                StackVarInfo theLocal = locals.get(i);
                int offset = firstLocalOffset - i * backend.getWordSize();
                localMap.put(theLocal.getVarName(), offset);
            }
        }
        public void pushLocalVars(List<StackVarInfo> locals) {
            // This function generates some asm code, so it should only be called at RIGHT time.
            for (int i = 0; i < locals.size(); ++i) {
                StackVarInfo theLocal = locals.get(i);
                theLocal.getInitialValue().dispatch(this);
                // The literal initial value is pushed to stack. That's the local!
                // No need to push it again.
            }
        }
        private void emitLoadNestedParentVar(StackVarInfo varInfo, RiscVBackend.Register tmpReg, RiscVBackend.Register valReg_to_assign) {
            // To assign: set valReg properly.
            // To read the var: set valReg to null. Returns valReg.
            FuncInfo hisOwner = varInfo.getFuncInfo();

            if (!mapFuncInfoToStmtAnalyzer.containsKey(hisOwner)) {
                new StmtAnalyzer(hisOwner);
            }

            //! assert: hisOwner is parent_of(this)
            FuncInfo curr = this.funcInfo;
            int counter = 0;
            while(curr != hisOwner && curr != null) {
                curr = curr.getParentFuncInfo();
                ++counter;
            }
            
            //! assert end
            assert curr != null;
            assert counter == funcInfo.getDepth() - hisOwner.getDepth();
            
            int depthDiff = funcInfo.getDepth() - hisOwner.getDepth();
            // If depthDiff is 0, then the variable is in current frame.
            backend.emitMV(tmpReg, FP, "LOAD NESTED VAR: Load current FP.");
            for (int i = 0; i < depthDiff; ++i) {
                backend.emitLW(tmpReg, tmpReg, 0 * backend.getWordSize(), "LOAD NESTED VAR: Load LOGICAL parent FP.");
            }
            int offsetInParent = mapFuncInfoToStmtAnalyzer.get(hisOwner).localMap.get(varInfo.getVarName());
            // Read the value to tmpReg
            if (valReg_to_assign == null)
                backend.emitLW(tmpReg, tmpReg, offsetInParent, "LOAD NESTED VAR: Read val: Set the result!");
            // Write to value:
            else
                backend.emitSW(valReg_to_assign, tmpReg, offsetInParent, "LOAD NESTED VAR: Write val: Set the val!");
        }

        /** An analyzer for the function described by FUNCINFO0, which is null
         *  for the top level. */
        StmtAnalyzer(FuncInfo funcInfo0) {
            funcInfo = funcInfo0;
            if (funcInfo == null)
                sym = globalSymbols;
            else
                sym = funcInfo.getSymbolTable();

            if (funcInfo != null) {
                // The init function below won't generate any asm code, so the stmtAnalyzer can be constructed at any time.
                // because the constructor doesn't generate code! It won't affect anything.
                initArgsMap(funcInfo.getParams());
                initLocalVars(funcInfo.getLocals(), -3 * backend.getWordSize());
                mapFuncInfoToStmtAnalyzer.put(funcInfo, this);
            }
        }

        @Override
        public Void analyze(ExprStmt exprStmt) {
            exprStmt.expr.dispatch(this);
            return null;
        }

        @Override
        public Void analyze(AssignStmt node) {
            node.value.dispatch(this);
            RiscVBackend.Register tmpReg = T6; // WARNING: May be overwritten!!!
            RiscVBackend.Register tmpReg2 = T5;
            betterBackend.emitPop(tmpReg, "ASSIGN: Get the value to assign: Pop the expr result from stack");

            SymbolType symbolType = null;
            for (Expr expr : node.targets) {
                if (expr instanceof MemberExpr) {
                    betterBackend.emitPush(tmpReg, "Push value to save during object dispatch");
                    ((MemberExpr) expr).object.dispatch(this);
                    betterBackend.emitPop(tmpReg2, "Load object");
                    Label normLabel = generateLocalLabel();
                    backend.emitBNEZ(tmpReg2, normLabel, "CHECK: Object is non-null");
                    backend.emitJAL(errorNone, "null error");
                    backend.emitLocalLabel(normLabel, "function normally");
                    betterBackend.emitPop(tmpReg, "Load value from stack");
                    MemberExpr memberExpr = (MemberExpr) expr;
                    ClassInfo classInfo = (ClassInfo) sym.get(memberExpr.object.getInferredType().className());
                    boolean foundAttr = false;
                    int attrCtr = 0;
                    for (AttrInfo a : classInfo.getAttributes()) {
                        if (a.getVarName().equals(memberExpr.member.name)) {
                            foundAttr = true;
                            break;
                        }
                        attrCtr++;
                    }
                    boolean foundMethod = false;
                    int methodCtr = 0;
                    if (!foundAttr) {
                        for (FuncInfo m : classInfo.getMethods()) {
                            if (m.getFuncName().equals(memberExpr.member.name)) {
                                foundMethod = true;
                                break;
                            }
                            methodCtr++;
                        }
                    }
                    if (foundAttr) {
                        int offset = attrCtr * backend.getWordSize() + 3 * backend.getWordSize();
                        backend.emitSW(tmpReg, tmpReg2, offset, "Assign object attr");
                    }
                }
                else if (expr instanceof IndexExpr) {
                    IndexExpr indexExpr = (IndexExpr) expr;
                    betterBackend.emitPush(tmpReg, "backup value to assign");

                    RiscVBackend.Register indexReg = A6, listReg = tmpReg2, lenReg = A7, elementReg = tmpReg;
                    indexExpr.list.dispatch(this);
                    indexExpr.index.dispatch(this);
                    betterBackend.emitPop(indexReg, "Assign list element: get index");
                    betterBackend.emitPop(listReg, "Assign list element: get list");

                    betterBackend.emitPop(tmpReg, "restore value to assign");

                    Label normLabel = generateLocalLabel();
                    backend.emitBNEZ(listReg, normLabel, "Assign list element: Exception: NullPointerException");
                    backend.emitJAL(errorNone, "null error");
                    backend.emitLocalLabel(normLabel, "function normally");
                    emitSetListElement(listReg, indexReg, lenReg, elementReg);
                }
                else {
                    SymbolInfo symbolInfo = sym.get(((Identifier) expr).name);

                    // do assign
                    if (symbolInfo instanceof GlobalVarInfo) {
                        GlobalVarInfo globalVarInfo = (GlobalVarInfo) symbolInfo;
                        backend.emitSW(tmpReg, globalVarInfo.getLabel(), tmpReg2, "ASSIGN: Assign to global var");
                    } else if (symbolInfo instanceof StackVarInfo) {
                        StackVarInfo stackVarInfo = (StackVarInfo) symbolInfo;
                        this.emitLoadNestedParentVar(stackVarInfo, tmpReg2, tmpReg); // Also work for local variable or param.
                    }
                }
            }
            return null;
        }

        private String getLabelPrefix() {
            String labelPrefix = "";
            if (funcInfo != null) {
                labelPrefix += "$" + funcInfo.getFuncName();
            }
            labelPrefix += "$__rlib$" + _r_local_label_counter;
            ++_r_local_label_counter;
            return labelPrefix;
        }

        @Override
        public Void analyze(IfStmt node) {
            node.condition.dispatch(this);
            RiscVBackend.Register tmpReg = T6;
            betterBackend.emitPop(tmpReg, "IfStmt: Get condition");

            String labelPrefix = getLabelPrefix();

            Label elseLabel = new Label( labelPrefix + ".ifstmt.else");
            Label exitLabel = new Label( labelPrefix + ".ifstmt.exit");
            backend.emitBEQ(ZERO, tmpReg, elseLabel, "IfStmt: If condition is false, jump to else.");
            for (Stmt stmt : node.thenBody) {
                stmt.dispatch(this);
            }
            backend.emitJ(exitLabel, "IfStmt: THEN body done. jump to end.");
            backend.emitLocalLabel(elseLabel, "IfStmt: ELSE");
            for (Stmt stmt : node.elseBody) {
                stmt.dispatch(this);
            }
            backend.emitLocalLabel(exitLabel, "IfStmt: DONE");

            return null;
        }

        @Override
        public Void analyze(IfExpr node) {
            node.condition.dispatch(this);
            RiscVBackend.Register tmpReg = T6;
            betterBackend.emitPop(tmpReg, "IfExpr: Get condition");

            String labelPrefix = getLabelPrefix();

            Label elseLabel = new Label( labelPrefix + ".ifexpr.else");
            Label exitLabel = new Label( labelPrefix + ".ifexpr.exit");
            backend.emitBEQ(ZERO, tmpReg, elseLabel, "IfExpr: If condition is false, jump to else.");
            node.thenExpr.dispatch(this);
            backend.emitJ(exitLabel, "IfExpr: THEN body done. jump to end.");
            backend.emitLocalLabel(elseLabel, "IfExpr: ELSE");
            node.elseExpr.dispatch(this);
            backend.emitLocalLabel(exitLabel, "IfExpr: DONE");

            return null;
        }

        @Override
        public Void analyze(WhileStmt node) {
            String labelPrefix = getLabelPrefix();

            Label beginLabel = new Label(labelPrefix + ".while.begin");
            Label exitLabel = new Label(labelPrefix + ".while.exit");

            backend.emitLocalLabel(beginLabel, "WhileStmt: Begin");
            node.condition.dispatch(this);
            RiscVBackend.Register tmpReg = T6;
            betterBackend.emitPop(tmpReg, "GetCondition");
            backend.emitBEQ(ZERO, tmpReg, exitLabel, "WhileStmt, IfFalse, exit");

            for (Stmt stmt : node.body) {
                stmt.dispatch(this);
            }
            backend.emitJ(beginLabel, "WhileStmt: GOTO begin");

            backend.emitLocalLabel(exitLabel, "WhileStmt: End");
            return null;
        }

        @Override
        public Void analyze(ForStmt node) {
            Label beginFor = new Label(getLabelPrefix() + ".forBegin");
            Label exitForDone = new Label(getLabelPrefix() + "exitForDone");
            Label exitForEmpty = new Label(getLabelPrefix() + "exitForEmpty");
            node.iterable.dispatch(this);
            SymbolType iType;
            if (node.iterable.getInferredType() instanceof ListValueType) {
                iType = node.iterable.getInferredType();
                betterBackend.emitPop(T1, "ForStmt: Get iterable");

                // check null
                Label normLabel = generateLocalLabel();
                backend.emitBNEZ(T1, normLabel, "CHECK: non null");
                backend.emitJAL(errorNone, "null error");
                backend.emitLocalLabel(normLabel, "function normally");

                // get length -> T2
                backend.emitLW(T2, T1, 3 * backend.getWordSize(), "ForStmt: Get list length");
                
                // if len == 0, jump to exitFor
                backend.emitBEQZ(T2, exitForEmpty, "Len == 0, jump to end");

                // push
                betterBackend.emitPush(T1, "ForStmt: Push list ptr to stack");
                betterBackend.emitPush(T2, "ForStmt: Push list length to stack");
            } else {
                iType = SymbolType.STR_TYPE;
                betterBackend.emitPop(T1, "ForStmt: Get string");
                backend.emitLW(T2, T1, 3 * backend.getWordSize(), "ForStmt: Get list length");
                
                // make sure str is not empty
                backend.emitBEQZ(T2, exitForEmpty, "CHECK: str len == 0");
                
                betterBackend.emitPush(T1, "ForStmt: Push list ptr to stack");
                betterBackend.emitPush(T2, "ForStmt: Push list length to stack");
            }
            SymbolInfo idInfo = sym.get(node.identifier.name);

            // idx reg prologue
            backend.emitLI(A6, -1, "ForStmt: set idx to -1");
            betterBackend.emitPush(A6, "ForStmt: Push idx to stack");
            // BEGIN FOR
            backend.emitLocalLabel(beginFor, "ForStmt: Begin for");
            // increment idx, check idx < length, if
            betterBackend.emitPop(A6, "ForStmt: Get idx from stack");
            backend.emitADDI(A6, A6, 1, "ForStmt: Increment idx");
            betterBackend.emitPop(A5, "ForStmt: Get length from stack");
            backend.emitBGEU(A6, A5, exitForDone, "ForStmt: if idx >= len go to exitFor");
            backend.emitLW(A4, SP, 0, "ForStmt: Peek list ptr");
            betterBackend.emitPush(A5, "ForStmt: Push length to stack, PRE FOR BODY");
            betterBackend.emitPush(A6, "ForStmt: Push idx back to stack, PRE FOR BODY");
            // backend.emitLW(T4, T4, 0, "ForStmt: load element");
            emitGetElementByIndex(A4, A6, A5, iType);
            // only peeked list ptr. still at top of stack --  betterBackend.emitPush(T3, "ForStmt: Push list ptr back to stack");

            // node.identifier should have value iterable[idx]
            if (idInfo instanceof GlobalVarInfo) {
                GlobalVarInfo idGlobalinfo = (GlobalVarInfo) idInfo;
                backend.emitSW(A4, idGlobalinfo.getLabel(), T5,"ForStmt: Assign global");
            } else {
                StackVarInfo idStackInfo = (StackVarInfo) idInfo;
                this.emitLoadNestedParentVar(idStackInfo, T5, A4);
            }
            
            // execute body
            for (Stmt s : node.body) {
                s.dispatch(this);
            }
            backend.emitJ(beginFor, "ForStmt: Jump to beginFor");
            backend.emitLocalLabel(exitForDone, "ForStmt: Exit for");
            // at exitFor, we've already gotten the idx and length from stack, so just need to decrement for listPtr
            betterBackend.emitPop(null, "ForStmt: Decrement stack for list ptr");
            backend.emitLocalLabel(exitForEmpty, "ForStmt: Exit for from empty or after decrement for done");
            return null;
        }

        @Override
        public Void analyze(ReturnStmt node) {
            if (node.value != null) {
                node.value.dispatch(this);
                betterBackend.emitPop(A0, "ReturnStmt: Get value");
            }
            // PEEK instead??
            // backend.emitLW(A0, SP, 0, "ReturnStmt: Peek value");
            betterBackend.emitFunctionReturn(this.funcInfo.getFuncName(), "ReturnStmt: GO");
            return null;
        }

        @Override
        public Void analyze(CallExpr node) {
            SymbolInfo called = sym.get(node.function.name);

            if (called instanceof FuncInfo) {
                FuncInfo func = (FuncInfo) called;
                boolean isBuiltIn = func.getFuncName().equals("print") || func.getFuncName().equals("len");

                for (Expr expr : node.args) {
                    expr.dispatch(this);
                    // ALL EXPR SHOULD NEVER EDIT STACK POINTER
                    // THE EXPR PUSH ITS RETURN VAL TO STACK, AS FUNCTION ARG

                    if (func.getFuncName().equals("print")) {
                        // FUNCTION PRINT.
                        if (expr.getInferredType().equals(SymbolType.INT_TYPE) || expr.getInferredType().equals(SymbolType.BOOL_TYPE))
                            betterBackend.emitWrapIntValToIntObject(T6, expr.getInferredType());
                    }
                }
                betterBackend.emitCall(func.getCodeLabel(), null, func == funcInfo, isBuiltIn);

                for (Object arg : node.args)
                    betterBackend.emitPop(null, "Pop function arguments.");

                if (! func.getFuncName().equals("print")) betterBackend.emitPush(A0, "CallExpr: Push function return val");
                return null;
            } else if (called instanceof ClassInfo) {
                // Implements constructors
                ClassInfo classInfo = (ClassInfo) called;

                backend.emitLA(A0, classInfo.getPrototypeLabel(), "Load prototype address");
                betterBackend.emitCall(objectAllocLabel, "Allocate new object", false, true);
                betterBackend.emitPush(A0, "Push new allocated addr on stack");

                for (Expr expr : node.args) {
                    expr.dispatch(this);
                }
                Label dispatchTable = classInfo.getDispatchTableLabel();
                backend.emitLA(T0, dispatchTable, "Load " + node.function.name + " dispatch table address");

                // load address of init method
                backend.emitLW(T0, T0, 0, "Load method address");
                betterBackend.emitCallR(T0, "Invoke ClassInfo method", false, false);
                betterBackend.emitPop(T1, "Pop the new addr off stack");
                for (Expr expr : node.args) {
                    betterBackend.emitPop(null, "Pop function arg");
                }
                // for int and bool, unbox the int val, for print function convention.
                if (classInfo.getClassName().equals("int") ||
                        classInfo.getClassName().equals("bool")) {
                    backend.emitLW(T1, T1, "@.__int__", "Load int value after constructor");
                }
                betterBackend.emitPush(T1, "Push the new addr back on stack");
                // betterBackend.emitPush(A0, "CallExpr: Push return val to stack");
            }
            return null;
        }

        @Override
        public Void analyze(MethodCallExpr methodCallExpr) {
            // push self to stack
            methodCallExpr.method.dispatch(this);
            betterBackend.emitPop(T4, "get method addr on stack");
            for (Expr expr : methodCallExpr.args) {
                expr.dispatch(this);
            }
            betterBackend.emitCallR(T4, "Invoke method", false, false);
            // WARN: Assume the method is not recursive
            for (Expr expr : methodCallExpr.args) {
                betterBackend.emitPop(null, "Pop Function Arg");
            }
            betterBackend.emitPush(A0, "Push return value");
            return null;
        }

        @Override
        public Void analyze(MemberExpr memberExpr) {
            memberExpr.object.dispatch(this);
            backend.emitLW(T1, SP, 0, "Peek member object"); 
            
            Label normLabel = generateLocalLabel();
            backend.emitBNEZ(T1, normLabel, "CHECK: Object is non-null");
            backend.emitJAL(errorNone, "null error");
            backend.emitLocalLabel(normLabel, "function normally");
            // calculate offset...
            String className = memberExpr.object.getInferredType().className();
            ClassInfo classInfo = (ClassInfo) sym.get(className);
            boolean foundAttr = false;
            int attrCtr = 0;
            for (AttrInfo a : classInfo.getAttributes()) {
                if (a.getVarName().equals(memberExpr.member.name)) {
                    foundAttr = true;
                    break;
                }
                attrCtr++;
            }
            boolean foundMethod = false;
            int methodCtr = 0;
            if (!foundAttr) {
                for (FuncInfo m : classInfo.getMethods()) {
                    if (m.getBaseName().equals(memberExpr.member.name)) {
                        foundMethod = true;
                        break;
                    }
                    methodCtr++;
                }
            }

            if (foundAttr) {
                // code attributes
                int offset = attrCtr * backend.getWordSize() + 3 * backend.getWordSize();
                backend.emitLW(T1, T1, offset, "Get attribute");
                betterBackend.emitPush(T1, "Push attribute to stack");
            } else if (foundMethod) {
                // code methods
                int offset = methodCtr * backend.getWordSize();
                Label dispatchTable = classInfo.getDispatchTableLabel();
                backend.emitLA(T1, dispatchTable, "Load object dispatch table");
                backend.emitLW(T1, T1, offset, "Load method address");
                betterBackend.emitPush(T1, "Push method address on stack");
            }
            return null;
        }

        @Override
        public Void analyze(NoneLiteral literal) {
            backend.emitMV(A0, ZERO, "NoneLiteral: load None (zero) into A0");
            betterBackend.emitPush(A0, "Push NoneLiteral onto stack");
            return null;
        }

        @Override
        public Void analyze(BooleanLiteral literal) {
            backend.emitLI(A0, (literal.value == true) ? 1 : 0, "Load boolean literal into a0");
            betterBackend.emitPush(A0, "Push literal onto stack");
            return null;
        }

        @Override
        public Void analyze(IntegerLiteral literal) {
            backend.emitLI(A0, literal.value, "Load integer literal into a0");
            betterBackend.emitPush(A0, "Push literal onto stack");
            return null;
        }

        @Override
        public Void analyze(StringLiteral literal) {
            backend.emitLA(A0, constants.getStrConstant(literal.value), "Load string addr into a0");
            betterBackend.emitPush(A0, "Push literal onto stack");
            return null;
        }

        @Override
        public Void analyze(Identifier id) {
            SymbolInfo symbolInfo = sym.get(id.name);
            RiscVBackend.Register tmpReg = T6;
            if (symbolInfo instanceof GlobalVarInfo) {
                backend.emitLW(tmpReg, ((GlobalVarInfo) symbolInfo).getLabel(), "Load identifier");
            } else if (symbolInfo instanceof StackVarInfo) {
                StackVarInfo stackVarInfo = (StackVarInfo) symbolInfo;
                this.emitLoadNestedParentVar(stackVarInfo, tmpReg, null); // Also work for local variable or param.
            }

            // Check if object is non-null
            // Label nonError = generateLocalLabel();
            /* NOTE: Commenting the following NULL check out for now. Moving it to outer methods. 
                        if (id.getInferredType() instanceof ClassValueType
                        && !id.getInferredType().equals(SymbolType.INT_TYPE)
                        && !id.getInferredType().equals(SymbolType.BOOL_TYPE)) {
                            backend.emitBEQZ(tmpReg, errorNone, "CHECK: Object is non-null");
                            //backend.emitJ(errorNone, "Jump to None error handler");
                            //backend.emitLocalLabel(nonError, "Object is non-null");
                        }
            */
            if (symbolInfo instanceof GlobalVarInfo)
                betterBackend.emitPush(tmpReg, "Ident: GlobalVar");
            else if (symbolInfo instanceof StackVarInfo)
                betterBackend.emitPush(tmpReg, "Ident: Arg or stack var");
            return null;
        }

        private void __rlib_emitStringLengthToWords(RiscVBackend.Register tmpReg) {
                 // I should do `tmp1 = ceil(tmp1 / WORD_SIZE)`
                backend.emitADDI(tmpReg, tmpReg, backend.getWordSize() - 1, "STR ADD: tmp1 = ceil(tmp1 / WORD_SIZE)");
                assert backend.getWordSize() == 4; // the following stmt requires it.
                backend.emitSRAI(tmpReg, tmpReg, 2, "STR ADD: tmp1 = ceil(tmp1 / WORD_SIZE)");
        }

        private RiscVBackend.Register emitStringAdd(RiscVBackend.Register left_const, RiscVBackend.Register right_const, boolean isStringElseList) {
            /* Assume already using: T5, T6 */ // isStringElseList: true for string, false for list.
            String prefix = isStringElseList ? "STR ADD: " : "LIST ADD: ";

            // Check for None
            Label normLabel = generateLocalLabel();
            backend.emitBNEZ(left_const, normLabel, prefix + "Exception: NullPointerException left");
            backend.emitBNEZ(right_const, normLabel, prefix + "Exception: NullPointerException right");
            backend.emitJAL(errorNone, "null error");
            backend.emitLocalLabel(normLabel, "function normally");

            // Prepare new object
            RiscVBackend.Register tmp1 = T4, tmp2 = A5;
            backend.emitLW(tmp1, left_const, 3*backend.getWordSize(), prefix + "get left len");
            backend.emitLW(tmp2, right_const, 3*backend.getWordSize(), prefix + "get right len");
            backend.emitADD(tmp1, tmp1, tmp2, prefix + "get lenLeft + lenRight");
            backend.emitMV(tmp2, tmp1, prefix + "backup previous result");

            if (isStringElseList)
                __rlib_emitStringLengthToWords(tmp1);

            // tmp1 is new string object size in word.
            backend.emitADDI(tmp1, tmp1, 4, prefix + "Calc len: Add header size.");
            if (isStringElseList)
                backend.emitADDI(tmp1, tmp1, 1, prefix + "STR BUG WORKAROUND: Off-by-1 bug. just alloc one more byte");
            // There should be some buffer-overflow for string! Don't worry, I alloc one more byte to contain junks!!!

            // alloc new string
            betterBackend.emitAllocPrototype(SymbolType.STR_TYPE, tmp1); // @returns A0
            backend.emitSW(tmp2, A0, 3 * backend.getWordSize(), prefix + "Set new string len");

            // tmp1 tmp2 die here.

            RiscVBackend.Register tmpNewStrDataPtr = A7, tmpOldStrDataPtr = A6;

            // calc left string words to copy
            backend.emitLW(tmp1, left_const, 3*backend.getWordSize(), prefix + "leftStr.len");
            if (isStringElseList)
                __rlib_emitStringLengthToWords(tmp1);

            // Copy left string to dest.
            backend.emitADDI(tmpNewStrDataPtr, A0, 4*backend.getWordSize(), prefix + "newStr.data");
            backend.emitADDI(tmpOldStrDataPtr, left_const, 4*backend.getWordSize(), prefix + "leftStr.data");
            betterBackend.emitMemcpy(tmpNewStrDataPtr, tmpOldStrDataPtr, tmp1, tmp2, getLabelPrefix());

            // calc right string words to copy
            backend.emitLW(tmp1, right_const, 3*backend.getWordSize(), prefix + "rightStr.len");
            if (isStringElseList)
                __rlib_emitStringLengthToWords(tmp1);

            // Calc memcpy arg: dstPtr.
            backend.emitLW(tmp2, left_const, 3*backend.getWordSize(), prefix + "leftStr.len");
            if (!isStringElseList) // for list: should add WORD_SIZE*leftStr.len
                backend.emitSLLI(tmp2, tmp2, 2, prefix + "leftStr.len *= 4 for list.");
            backend.emitADDI(tmpNewStrDataPtr, A0, 4*backend.getWordSize(), prefix + "newStr.data");
            backend.emitADD(tmpNewStrDataPtr, tmpNewStrDataPtr, tmp2, prefix + "&newStr.data + leftStr.len");
            
            // Copy right string to dest.
            backend.emitADDI(tmpOldStrDataPtr, right_const, 4*backend.getWordSize(), prefix + "rightStr.data");
            betterBackend.emitMemcpy(tmpNewStrDataPtr, tmpOldStrDataPtr, tmp1, tmp2, getLabelPrefix());
            
            // There should be some buffer-overflow for string! Don't worry, I alloc one more byte to contain junks!!!
            return A0;
        }

        @Override
        public Void analyze(BinaryExpr binaryExpr) {
            SymbolType leftType = binaryExpr.left.getInferredType();
            SymbolType rightType = binaryExpr.right.getInferredType();
            String op = binaryExpr.operator;
            RiscVBackend.Register leftReg = T6, rightReg = T5;

            RiscVBackend.Register resultReg = leftReg;

            if (!(op.equals("and") || op.equals("or"))) {
                binaryExpr.left.dispatch(this);
                binaryExpr.right.dispatch(this);
                betterBackend.emitPop(rightReg, "Load right operand");
                betterBackend.emitPop(leftReg, "Load left operand");
            }

            switch (binaryExpr.operator) {
                case "+":
                    if (leftType.equals(SymbolType.INT_TYPE))
                        backend.emitADD(resultReg, leftReg, rightReg, "Add operands");
                    else if (leftType.equals(SymbolType.STR_TYPE) || leftType.isListType() || leftType.equals(SymbolType.EMPTY_TYPE))
                        resultReg = emitStringAdd(leftReg, rightReg, leftType.equals(SymbolType.STR_TYPE));
                    break;
                case "-":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE))   
                        backend.emitSUB(resultReg, leftReg, rightReg, "Subtract operands");
                    break;
                case "*":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) 
                        backend.emitMUL(resultReg, leftReg, rightReg, "Subtract operands");
                    break;
                case "//":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {    
                        Label retLabel = generateLocalLabel();
                        backend.emitBNEZ(rightReg, retLabel, "Check if operand 2 is Zero");
                        backend.emitJAL(errorDiv, "div error");
                        backend.emitLocalLabel(retLabel, "function normally");
                        backend.emitDIV(resultReg, leftReg, rightReg, "Divide operands");
                    }
                    break;
                case "%":
                    if (binaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {    
                        Label retLabel = generateLocalLabel();
                        backend.emitBNEZ(rightReg, retLabel, "Check if operand 2 is Zero");
                        backend.emitJAL(errorDiv, "div error");
                        backend.emitLocalLabel(retLabel, "function normally");
                        backend.emitREM(resultReg, leftReg, rightReg, "Modulus operands");
                    }
                    break;
                case "==":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE) ||
                            binaryExpr.getInferredType().equals(SymbolType.INT_TYPE) ) {
                        backend.emitXOR(leftReg, rightReg, leftReg, "== operator on BOOL/INT");
                        backend.emitSEQZ(resultReg, leftReg, "Set to 1 if XOR results in 0, they were equal");
                    }

                    break;
                case "!=":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE) ||
                            binaryExpr.getInferredType().equals(SymbolType.INT_TYPE) ) {
                        backend.emitXOR(resultReg, rightReg, leftReg, "!= operator on BOOL/INT");
                        // backend.emitSNEZ(resultReg, resultReg, "Set to 1 if XOR was not 0, they were not equal");
                        // Don't think we need to do the SEQZ/SNEZ instructions here because XOR will already result in 1
                        // if they are not equal, which is the desired result.
                    }
                    break;
                case "<=":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        backend.emitADDI(leftReg, leftReg, -1, "decrement LEFT by 1");
                        backend.emitSLT(resultReg, leftReg, rightReg, "LEFT <=? RIGHT");
                    }
                    break;
                case ">=":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE)) {
                        backend.emitADDI(rightReg, rightReg, -1, "decrement RIGHT by 1");
                        backend.emitSLT(resultReg, rightReg, leftReg, "LEFT >=? RIGHT");
                    }
                    break;
                case "<":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE))
                        backend.emitSLT(resultReg, leftReg, rightReg, "LEFT <? RIGHT");
                    break;
                case ">":
                    if (binaryExpr.left.getInferredType().equals(SymbolType.INT_TYPE))
                        backend.emitSLT(resultReg, rightReg, leftReg, "LEFT >? RIGHT");
                    break;
                case "and":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        betterBackend.emitPop(resultReg, "Load operand 1 into t1");
                        
                        Label skipSecond = generateLocalLabel();
                        backend.emitBEQZ(resultReg, skipSecond, "If operand 1 is FALSE, don't eval second");
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(rightReg, "Load operand 2 into t0");
                        backend.emitAND(resultReg, resultReg, rightReg, "AND operands");
                        backend.emitLocalLabel(skipSecond, "After second operand");
                    }
                    break;
                case "or":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE)) {
                        binaryExpr.left.dispatch(this);
                        betterBackend.emitPop(resultReg, "Load operand 1 into t0");
                        
                        Label skipSecond = generateLocalLabel();
                        backend.emitBNEZ(resultReg, skipSecond, "If first is not zero (it's true), skip second");
                        binaryExpr.right.dispatch(this);
                        betterBackend.emitPop(rightReg, "Load operand 2 into t0");
                        backend.emitOR(resultReg, resultReg, rightReg, "OR operands");
                        backend.emitLocalLabel(skipSecond, "Skip past second operand");
                    }
                    break;
                case "is":
                    if (binaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE) ||
                            binaryExpr.getInferredType().equals(SymbolType.INT_TYPE) ) {

                        backend.emitXOR(rightReg, rightReg, leftReg, "is operator");
                        backend.emitSEQZ(resultReg, rightReg, "Set to 1 if XOR results in 0, they were equal");
                    }
                    break;
                default:
                    throw new RuntimeException("BinaryExpr: NOT IMPLEMENTED operator " + op);
            }

            betterBackend.emitPush(resultReg, "Push result onto stack");
            return null;
        }

        @Override
        public Void analyze(UnaryExpr unaryExpr) {
            switch (unaryExpr.operator) {
                case "-":
                    if (unaryExpr.getInferredType().equals(SymbolType.INT_TYPE)) {
                        unaryExpr.operand.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand into t0");
                        
                        backend.emitSUB(T0, ZERO, T0, "Subtract operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
                case "not":
                    if (unaryExpr.getInferredType().equals(SymbolType.BOOL_TYPE)) {
                        unaryExpr.operand.dispatch(this);
                        betterBackend.emitPop(T0, "Load operand into t0");
                        
                        backend.emitSEQZ(T0, T0, "NOT operands");
                        betterBackend.emitPush(T0, "Push result onto stack");
                        return null;
                    }
            }
            return null;
        }

        private void emitGetElementByIndex(RiscVBackend.Register listReg, RiscVBackend.Register indexReg, RiscVBackend.Register lenReg, SymbolType listType) {
            //! Note: All registers in argument will be modified! listType should be listType or STR_TYPE
            //! Note: You need to set listReg, indexReg. lenReg is just a temp reg, no need to set it.
            //! @returns: listReg
            //! listReg is resultReg. Return list element ptr, or new string ptr.
            backend.emitLW(lenReg, listReg, 3*backend.getWordSize(), "IndexExpr: Get List/str length");
            backend.emitADDI(listReg, listReg, 4*backend.getWordSize(), "IndexExpr: listPtr = list.data");

            String labelPrefix = getLabelPrefix();
            Label skipTranslateNegIndex = new Label(labelPrefix + ".NegIndexEnd");

            // Get real index.
            // Recolic: they don't like "abc"[-1], they want to crash it . Remove the feature and make them happy.
            // backend.emitBGEZ(indexReg, skipTranslateNegIndex, "IndexExpr: Skip neg index op");
            // backend.emitADD(indexReg, indexReg, lenReg, "IndexExpr: Add length to negative index.");
            backend.emitBGEZ(indexReg, skipTranslateNegIndex, "IndexExpr: Exception: negative index still negative after adding length.");
            backend.emitJAL(errorOob, "oob error");
            backend.emitLocalLabel(skipTranslateNegIndex, "IndexExpr: DO REAL ACCESS");
            
            // Do real access!
            Label normLabel= generateLocalLabel();
            backend.emitBLT(indexReg, lenReg, normLabel, "IndexExpr: Exception: index larger than length.");
            backend.emitJAL(errorOob, "oob error");
            backend.emitLocalLabel(normLabel, "function normally");
            
            if (!listType.equals(SymbolType.STR_TYPE)) {
                assert backend.getWordSize() == 4;
                backend.emitSLLI(indexReg, indexReg, 2, "IndexExpr: for list, element size = 4 char, so: indexReg *= 4");
            }
            backend.emitADD(listReg, listReg, indexReg, "IndexExpr: listPtr += index");
            backend.emitLW(listReg, listReg, 0, "IndexExpr: result = *listPtr (reuse listPtr)");
            // listReg,indexReg,lenReg die. listReg => elementReg.
            if (listType.equals(SymbolType.STR_TYPE)) {
                // lenReg => newStrLenReg
                backend.emitANDI(listReg, listReg, 0x000000FF, "IndexExpr: for string element, wipe higher 3 chars, leave only 1 char");
                backend.emitLI(lenReg, 1+4, "IndexExpr: result string object size = 4+1 = 5");

                betterBackend.emitAllocPrototype(SymbolType.STR_TYPE, lenReg);

                backend.emitLI(lenReg, 1, "IndexExpr: result string len = 1");
                backend.emitSW(lenReg, A0, 3*backend.getWordSize(), "IndexExpr: result string.len");
                backend.emitSW(listReg, A0, 4*backend.getWordSize(), "IndexExpr: result string.data");

                backend.emitMV(listReg, A0, "IndexExpr: move result string to resultReg");
            }
        }

         private void emitSetListElement(RiscVBackend.Register listReg, RiscVBackend.Register indexReg, RiscVBackend.Register lenReg, RiscVBackend.Register elementReg) {
            /* DO NOT edit this function. it's copy-pasted from emitGetElement. keep them consist. */
            //! Note: All registers in argument will be modified! listType should be listType or STR_TYPE
            //! Note: You need to set listReg, indexReg, elementReg. lenReg is just a temp reg, no need to set it.
            backend.emitLW(lenReg, listReg, 3*backend.getWordSize(), "IndexExpr: Get List/str length");
            backend.emitADDI(listReg, listReg, 4*backend.getWordSize(), "IndexExpr: listPtr = list.data");

            String labelPrefix = getLabelPrefix();
            Label skipTranslateNegIndex = new Label(labelPrefix + ".NegIndexEnd");

            // Get real index.
            // Recolic: they don't like "abc"[-1], they want to crash it . Remove the feature and make them happy.
            // backend.emitBGEZ(indexReg, skipTranslateNegIndex, "IndexExpr: Skip neg index op");
            // backend.emitADD(indexReg, indexReg, lenReg, "IndexExpr: Add length to negative index.");
            backend.emitBGEZ(indexReg, skipTranslateNegIndex, "IndexExpr: Exception: negative index still negative after adding length.");
            backend.emitJAL(errorOob, "oob error");
            backend.emitLocalLabel(skipTranslateNegIndex, "IndexExpr: DO REAL ACCESS");
            
            // Do real access!
            Label normLabel = generateLocalLabel();
            backend.emitBLT(indexReg, lenReg, normLabel, "IndexExpr: Exception: index larger than length.");
            backend.emitJAL(errorOob, "oob error");
            backend.emitLocalLabel(normLabel, "function normally");
            
            if (true) {
                assert backend.getWordSize() == 4;
                backend.emitSLLI(indexReg, indexReg, 2, "IndexExpr: for list, element size = 4 char, so: indexReg *= 4");
            }
            backend.emitADD(listReg, listReg, indexReg, "IndexExpr: listPtr += index");

            // do real set!
            backend.emitSW(elementReg, listReg, 0, "IndexExpr Assign: Do real assign.");
        }



        @Override
        public Void analyze(IndexExpr node) {
            node.list.dispatch(this);
            node.index.dispatch(this);
            RiscVBackend.Register listReg = T6, indexReg = T5, lenReg = A7;
            betterBackend.emitPop(indexReg, "IndexExpr: Get Index");
            betterBackend.emitPop(listReg, "IndexExpr: Get List/str ptr");

            Label normLabel = generateLocalLabel();
            backend.emitBNEZ(listReg, normLabel, "IndexExpr: Exception: NullPointerException");
            backend.emitJAL(errorNone, "null error");
            backend.emitLocalLabel(normLabel, "function normally");

            emitGetElementByIndex(listReg, indexReg, lenReg, node.list.getInferredType());
            // listReg => resultReg.

            betterBackend.emitPush(listReg, "IndexExpr: push result!");
            return null;
        }

        @Override
        public Void analyze(ListExpr node) {
            for (Expr expr : node.elements) {
                expr.dispatch(this);
            }
            int listSize = node.elements.size();
            RiscVBackend.Register listReg = A0, tmpReg = T5;
            backend.emitLI(tmpReg, listSize + 4, "ListExpr: sizeof list object");
            betterBackend.emitAllocPrototype(node.getInferredType(), tmpReg);
            // listReg is assigned now.

            backend.emitLI(tmpReg, listSize, "ListExpr: list.size");
            backend.emitSW(tmpReg, listReg, 3*backend.getWordSize(), "ListExpr: set list.size");

            for (int i = 0; i < listSize; ++i) {
                betterBackend.emitPop(tmpReg, "ListExpr: fetch element.");
                backend.emitSW(tmpReg, listReg, (4 + (listSize-1-i))*backend.getWordSize(), "ListExpr: assign element");
            }

            betterBackend.emitPush(listReg, "ListExpr: push final result");
            return null;
        }


    }

    /**
     * Emits custom code in the CODE segment.
     *
     * This method is called after emitting the top level and the
     * function bodies for each function.
     *
     * You can use this method to emit anything you want outside of the
     * top level or functions, e.g. custom routines that you may want to
     * call from within your code to do common tasks. This is not strictly
     * needed. You might not modify this at all and still complete
     * the assignment.
     *
     * To start you off, here is an implementation of three routines that
     * will be commonly needed from within the code you will generate
     * for statements.
     *
     * The routines are error handlers for operations on None, index out
     * of bounds, and division by zero. They never return to their caller.
     * Just jump to one of these routines to throw an error and
     * exit the program. For example, to throw an OOB error:
     *   backend.emitJ(errorOob, "Go to out-of-bounds error and abort");
     *
     */
    protected void emitCustomCode() {
//        backend.emitGlobalLabel(new Label("int-alloc"));
//        backend.emitLA(A0, intClass.getPrototypeLabel(), "Get int prototype address");
//        backend.emitJAL(objectAllocLabel, "Allocate int");
//
//        FuncInfo allocInt = makeFuncInfo("int-alloc", 0, SymbolType.OBJECT_TYPE,
//                globalSymbols, null, this::emitStdFunc);
//        allocInt.addParam(makeStackVarInfo("arg", SymbolType.OBJECT_TYPE,
//                null, allocInt));
//        functions.add(allocInt);
//        System.out.println(allocInt.getBaseName());
//        globalSymbols.put(allocInt.getBaseName(), allocInt);
        emitErrorFunc(errorNone, "Operation on None", ERROR_NONE);
        emitErrorFunc(errorDiv, "Divison by zero", ERROR_DIV_ZERO);
        emitErrorFunc(errorOob, "Index out of bounds", ERROR_OOB);
    }

    /** Emit an error routine labeled ERRLABEL that aborts with message MSG. */
    private void emitErrorFunc(Label errLabel, String msg, int err) {
        backend.emitGlobalLabel(errLabel);
        backend.emitLI(A0, err, "Exit code for: " + msg);
        backend.emitLA(A1, constants.getStrConstant(msg),
                       "Load error message as str");
        backend.emitADDI(A1, A1, getAttrOffset(strClass, "__str__"),
                         "Load address of attribute __str__");
        backend.emitJ(abortLabel, "Abort");
    }
}
